@@ -1,16 +1,17 @@
-import importlib.metadata
+import importlib_metadata
 
 from akerbp.mlpet import (
     feature_engineering,
     imputers,
     petrophysical_features,
+    plotting,
     preprocessors,
     utilities,
 )
 from akerbp.mlpet.dataset import Dataset
 from akerbp.mlpet.transformer import MLPetTransformer
 
-__version__ = importlib.metadata.version(__name__)
+__version__ = importlib_metadata.version(__name__)
 
 __all__ = [
     "Dataset",
@@ -20,4 +21,5 @@ __all__ = [
     "preprocessors",
     "petrophysical_features",
     "MLPetTransformer",
+    "plotting",
 ]
