@@ -1,10 +1,10 @@
 import os
 import warnings
+from collections.abc import Iterable
 from copy import deepcopy
 from pathlib import Path
 from typing import Any, Dict, List, Optional, Set, Union
 
-import numpy as np
 import pandas as pd
 import yaml
 from pandas.core.frame import DataFrame
@@ -137,7 +137,7 @@ class Dataset(dl.DataLoader):
                     "function in your preprocessing_pipeling!"
                 ) from e
 
-    def __standardize_curves(self) -> None:
+    def __standardize_curves(self) -> None:  # noqa: C901
         """
         A helper function to standardize curve names.
 
@@ -151,6 +151,7 @@ class Dataset(dl.DataLoader):
             "curves_to_drop",
             "curves_to_impute",
             "columns_to_encode",
+            "columns_to_onehot_encode",
             "rolling_features",
             "gradient_features",
             "log_features",
@@ -161,6 +162,7 @@ class Dataset(dl.DataLoader):
             "numerical_curves",
             "categorical_curves",
             "keep_columns",
+            "columns_to_fill",
         ]
         all_curves = {}
         for curve_set in curve_sets:
@@ -169,16 +171,39 @@ class Dataset(dl.DataLoader):
 
         # Standardize passed curves if mappings exist
         if hasattr(self, "curve_mappings"):
-            for curve_set, curves in all_curves.items():
-                setattr(self, f"{curve_set}_original", curves)
-                new_names, _ = utilities.standardize_names(
-                    names=curves, mapper=self.curve_mappings
-                )
+            for curve_set, names in all_curves.items():
+                setattr(self, f"{curve_set}_original", names)
+                if isinstance(names, dict):
+                    new_names = {}
+                    for k, v in names.items():
+                        v.insert(0, k)
+                        new_v, _ = utilities.standardize_names(
+                            names=v, mapper=self.curve_mappings
+                        )
+                        new_names[new_v[0]] = new_v[1:]
+                else:
+                    new_names, _ = utilities.standardize_names(
+                        names=names, mapper=self.curve_mappings
+                    )
                 setattr(self, curve_set, new_names)
                 all_curves[curve_set] = new_names
 
         # Clean up all curves to be on one level and unique
-        self.all_curves = set(np.concatenate(list(all_curves.values())))
+        all = []
+        for _, v in all_curves.items():
+            if isinstance(v, dict):
+                all.extend(list(v.items()))
+            else:
+                all.extend(v)
+
+        def flatten(iterable):
+            for el in iterable:
+                if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
+                    yield from flatten(el)
+                else:
+                    yield el
+
+        self.all_curves = set(flatten(all))  # type: ignore
 
         # Standardize single curves if mappings exist
         curves = {"id_column": self.id_column}
@@ -251,6 +276,7 @@ class Dataset(dl.DataLoader):
                 feature_engineering.add_formations_and_groups,
                 feature_engineering.add_vertical_depths,
                 feature_engineering.add_petrophysical_features,
+                feature_engineering.add_depth_trend,
                 imputers.impute_depth_trend,
                 preprocessors.set_as_nan,
                 preprocessors.remove_outliers,
@@ -258,12 +284,14 @@ class Dataset(dl.DataLoader):
                 preprocessors.fill_zloc_from_depth,
                 preprocessors.fillna_with_fillers,
                 preprocessors.encode_columns,
+                preprocessors.onehot_encode_columns,
                 preprocessors.select_columns,
                 preprocessors.normalize_curves,
                 preprocessors.scale_curves,
                 preprocessors.process_wells,
                 preprocessors.remove_noise,
                 preprocessors.drop_columns,
+                preprocessors.fill_columns,
             ]
         }
         # <--------------------- INGEST INIT INPUTS -------------------------> #
@@ -279,8 +307,11 @@ class Dataset(dl.DataLoader):
             # Any keys in the provided mappings override the base mappings
             try:
                 curve_mappings = default_mappings.base_mappings["curve_mappings"]
-                curve_mappings.update(self.mappings["curve_mappings"])
-                self.mappings["curve_mappings"] = curve_mappings
+                curve_mappings.update()
+                self.mappings["curve_mappings"] = {
+                    **curve_mappings,
+                    **self.mappings["curve_mappings"],
+                }
             except KeyError:
                 self.mappings["curve_mappings"] = default_mappings.base_mappings[
                     "curve_mappings"
@@ -288,8 +319,10 @@ class Dataset(dl.DataLoader):
                 pass
             try:
                 formations_map = default_mappings.base_mappings["formations_map"]
-                formations_map.update(self.mappings["formations_map"])
-                self.mappings["formations_map"] = formations_map
+                self.mappings["formations_map"] = {
+                    **formations_map,
+                    **self.mappings["formations_map"],
+                }
             except KeyError:
                 self.mappings["formations_map"] = default_mappings.base_mappings[
                     "formations_map"
@@ -297,16 +330,30 @@ class Dataset(dl.DataLoader):
                 pass
             try:
                 groups_map = default_mappings.base_mappings["groups_map"]
-                groups_map.update(self.mappings["groups_map"])
-                self.mappings["groups_map"] = groups_map
+                self.mappings["groups_map"] = {
+                    **groups_map,
+                    **self.mappings["groups_map"],
+                }
             except KeyError:
                 self.mappings["groups_map"] = default_mappings.base_mappings[
                     "groups_map"
                 ]
                 pass
+            try:
+                systems_map = default_mappings.base_mappings["systems_map"]
+                self.mappings["systems_map"] = {
+                    **systems_map,
+                    **self.mappings["systems_map"],
+                }
+            except KeyError:
+                self.mappings["systems_map"] = default_mappings.base_mappings[
+                    "systems_map"
+                ]
+                pass
         self.curve_mappings = self.mappings["curve_mappings"]
         self.formations_map = self.mappings["formations_map"]
         self.groups_map = self.mappings["groups_map"]
+        self.systems_map = self.mappings["systems_map"]
 
         # Ensure required settings were provided to prevent problems later down the line
         required = ["id_column"]
@@ -353,7 +400,9 @@ class Dataset(dl.DataLoader):
         else:
             self.categorical_curves = [self.id_column]
 
-    def preprocess(self, df: DataFrame = None, verbose=True, **kwargs) -> DataFrame:
+    def preprocess(  # noqa: C901
+        self, df: Optional[DataFrame] = None, verbose=True, **kwargs
+    ) -> DataFrame:
         """
         Main preprocessing function. Pass the dataframe to be preprocessed along
         with any kwargs for running any desired order (within reason) of the
@@ -445,14 +494,18 @@ class Dataset(dl.DataLoader):
             # instantiation via the settings file. Taking a deepcopy because
             # we don't want to mutate the original pipeline with general defaults
             # in case it is to be used again later
-            if self.preprocessing_pipeline is not None:
-                kwargs = deepcopy(self.preprocessing_pipeline)
+            msg = (
+                "No preprocessing kwargs were passed (either at runtime or "
+                "via the settings file at instantiation). There's nothing "
+                "to preprocess!"
+            )
+            if hasattr(self, "preprocessing_pipeline"):
+                if self.preprocessing_pipeline is not None:
+                    kwargs = deepcopy(self.preprocessing_pipeline)
+                else:
+                    raise ValueError(msg)
             else:
-                raise ValueError(
-                    "No preprocessing kwargs were passed (either at runtime or "
-                    "via the settings file at instantiation). There's nothing "
-                    "to preprocess!"
-                )
+                raise ValueError(msg)
 
         # Fill in the blanks where necessary
         kwargs = self.get_preprocess_defaults(kwargs)
@@ -462,9 +515,21 @@ class Dataset(dl.DataLoader):
             kwargs.items(), desc="Preprocessing", disable=(not verbose), unit="function"
         )
         artifacts = {}
+        start_columns = set(df.columns)
+        new_features: Set[List[str]] = set([])
         for function, settings in pbar:
             if verbose:
                 tqdm.write(f"Running {function}")
+            if function == "scale_curves":
+                # Special case for scale_curves to reduce verbosity of settings files
+                if "scale_added_curves" in settings and new_features:
+                    if settings["curves_to_scale"] is None:
+                        settings["curves_to_scale"] = list(new_features)
+                    else:
+                        settings["curves_to_scale"] += list(new_features)
+                        settings["curves_to_scale"] = list(
+                            set(settings["curves_to_scale"])
+                        )
             try:
                 res = self.supported_preprocessing_functions[function](df, **settings)
             except Exception as e:
@@ -480,6 +545,8 @@ class Dataset(dl.DataLoader):
                 if isinstance(artifact, dict):
                     # safe to proceed with saving to cls
                     for k, v in artifact.items():
+                        # Note this is not safeguarded for potentially
+                        # overwriting existing attributes!
                         setattr(self, k, v)
                     artifacts.update(artifact)
                 else:
@@ -498,12 +565,11 @@ class Dataset(dl.DataLoader):
                 raise ValueError(
                     f"The preprocessing function {function} returned an illegal return type!"
                 )
+            new_features = set(df.columns) - start_columns
 
         # Perform admin work on detecting features created and removed and
         # artifacts created
-        self.features_added = list(
-            set([x for x in df.columns if x not in original_columns])
-        )
+        self.features_added = new_features
         self.original_columns_removed = list(
             set([x for x in original_columns if x not in df.columns])
         )
@@ -513,7 +579,7 @@ class Dataset(dl.DataLoader):
         return df
 
     def get_preprocess_defaults(
-        self, kwargs: Dict[str, Dict[str, Any]] = None
+        self, kwargs: Optional[Dict[str, Dict[str, Any]]] = None
     ) -> Dict[str, Any]:
         """
         Wrapper function to define and provide the default kwargs to use for
@@ -557,6 +623,17 @@ class Dataset(dl.DataLoader):
             "add_petrophysical_features": {
                 "petrophysical_features": getattr(self, "petrophysical_features", None),
                 "id_column": self.id_column,
+                "keyword_arguments": {  # VSH specific kwargs
+                    "nan_numerical_value": self.num_filler,
+                    "nan_textual_value": self.cat_filler,
+                },
+            },
+            "add_depth_trend": {
+                "id_column": self.id_column,
+                "keyword_arguments": {
+                    "nan_numerical_value": self.num_filler,
+                    "nan_textual_value": self.cat_filler,
+                },
             },
             "simple_impute": {
                 "categorical_curves": getattr(self, "categorical_curves", None),
@@ -604,11 +681,19 @@ class Dataset(dl.DataLoader):
                 ),
                 "formations_map": getattr(self, "formations_map", None),
                 "groups_map": getattr(self, "groups_map", None),
-                "missing_encoding_value": getattr(self, "missing_encoding_value", -1),
+                "systems_map": getattr(self, "systems_map", None),
+                "missing_encoding_value": getattr(
+                    self, "missing_encoding_value", self.num_filler
+                ),
+            },
+            "onehot_encode_columns": {
+                "columns_to_onehot_encode": getattr(
+                    self, "columns_to_onehot_encode", None
+                ),
             },
             "select_columns": {
                 "curves_to_select": getattr(self, "curves_to_select", None),
-                "label_column": self.label_column,
+                "label_column": getattr(self, "label_column", None),
                 "id_column": self.id_column,
             },
             "drop_columns": {
@@ -640,6 +725,10 @@ class Dataset(dl.DataLoader):
                 "noisy_curves": getattr(self, "noisy_curves", None),
                 "noise_removal_window": getattr(self, "noise_removal_window", None),
             },
+            "fill_columns": {
+                "columns_to_fill": getattr(self, "columns_to_fill", None),
+                "missing_value": getattr(self, "numerical_value", None),
+            },
         }
 
         # Process wells uses a bunch of lower level functions so we need to
@@ -675,6 +764,7 @@ class Dataset(dl.DataLoader):
                             "formations_map",
                             "groups_map",
                             "curves_mapping",
+                            "keyword_arguments",
                         ]:  # Append/Overwrite user provided mappings to existing mappings
                             kwargs[function_name][setting_name] = {
                                 **default_setting,
@@ -712,6 +802,17 @@ class Dataset(dl.DataLoader):
         if hasattr(self, "label_column"):
             expected_but_missing_curves -= set([self.label_column])
 
+        # Special case for depth column, if its iniitialised by doesn't exist in the dataframe
+        # we don't want to fill it with num_filler. Instead we should raise an error
+        if hasattr(self, "depth_column"):
+            if self.depth_column not in df.columns:
+                raise ValueError(
+                    f"Depth column {self.depth_column} was set in your settings"
+                    " file/preprocessing kwargs but could not be found in the "
+                    "provided dataframe. Cannot continue without a proper depth "
+                    "column."
+                )
+
         if expected_but_missing_curves:
             expected_but_missing_cat_curves = expected_but_missing_curves & set(
                 self.categorical_curves
