@@ -67,7 +67,7 @@ def add_rolling_features(
     **kwargs,
 ) -> pd.DataFrame:
     """
-    Creates columns with window/rolling features of curves. All created columns
+    Creates columns with centered window/rolling features of curves. All created columns
     are suffixed with '_window_mean' / '_window_max' / '_window_min'.
 
     Args:
@@ -81,6 +81,12 @@ def add_rolling_features(
             features. **The window size is defined in distance**! The sampling rate
             is determined from the depth_column kwarg and used to transform the window
             size into an index based window. If this is not provided, no rolling features are calculated.
+        calculate_mean (bool): Whether to calculate the mean of the window. Defaults to True.
+        calculate_max (bool): Whether to calculate the max of the window. Defaults to True.
+        calculate_min (bool): Whether to calculate the min of the window. Defaults to True.
+        calculate_var (bool): Whether to calculate the variance of the window. Defaults to False.
+        calculate_norm_dist (bool): Whether to calculate the normalized distance the current point is from the window min and max. Defaults to False.
+            calculate_min and calculate_max must be True for this to work.
 
     Returns:
         pd.DataFrame: New dataframe with calculated rolling feature columns
@@ -88,27 +94,54 @@ def add_rolling_features(
     rolling_features: List[str] = kwargs.get("rolling_features", None)
     window = kwargs.get("window", None)
     depth_column = kwargs.get("depth_column", None)
+    calculate_mean = kwargs.get("calculate_mean", True)
+    calculate_max = kwargs.get("calculate_max", True)
+    calculate_min = kwargs.get("calculate_min", True)
+    calculate_var = kwargs.get("calculate_var", False)
+    calculate_norm_dist = kwargs.get("calculate_norm_dist", False)
     if rolling_features is not None and window is not None and depth_column is not None:
+        curves_to_drop = []
         sampling_rate = utilities.calculate_sampling_rate(df[depth_column])
         window_size = int(window / sampling_rate)
-        mean_cols = [col + "_window_mean" for col in rolling_features]
-        df[mean_cols] = (
-            df[rolling_features]
-            .rolling(center=False, window=window_size, min_periods=1)
-            .mean()
-        )
-        min_cols = [col + "_window_min" for col in rolling_features]
-        df[min_cols] = (
-            df[rolling_features]
-            .rolling(center=False, window=window_size, min_periods=1)
-            .min()
-        )
-        max_cols = [col + "_window_max" for col in rolling_features]
-        df[max_cols] = (
-            df[rolling_features]
-            .rolling(center=False, window=window_size, min_periods=1)
-            .max()
-        )
+        if calculate_mean:
+            mean_cols = [col + "_window_mean" for col in rolling_features]
+            df[mean_cols] = (
+                df[rolling_features]
+                .rolling(center=True, window=window_size, min_periods=1)
+                .mean()
+            )
+        if calculate_min or calculate_norm_dist:
+            min_cols = [col + "_window_min" for col in rolling_features]
+            df[min_cols] = (
+                df[rolling_features]
+                .rolling(center=True, window=window_size, min_periods=1)
+                .min()
+            )
+            if not calculate_min:
+                curves_to_drop.extend(min_cols)
+        if calculate_max or calculate_norm_dist:
+            max_cols = [col + "_window_max" for col in rolling_features]
+            df[max_cols] = (
+                df[rolling_features]
+                .rolling(center=True, window=window_size, min_periods=1)
+                .max()
+            )
+            if not calculate_max:
+                curves_to_drop.extend(max_cols)
+        if calculate_var:
+            var_cols = [col + "_window_var" for col in rolling_features]
+            df[var_cols] = (
+                df[rolling_features]
+                .rolling(center=True, window=window_size, min_periods=1)
+                .var()
+            )
+        if calculate_norm_dist:
+            for col in rolling_features:
+                df[col + "_window_norm_dist"] = (df[col] - df[col + "_window_min"]) / (
+                    df[col + "_window_max"] - df[col + "_window_min"]
+                )
+        if curves_to_drop:
+            df = df.drop(columns=curves_to_drop, errors="ignore")
     return df
 
 
@@ -159,6 +192,7 @@ def add_petrophysical_features(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
             - BS is calculated using the guess_BS_from_CALI function from this
             module it is not found in the pass dataframe
         - VSH = Refer to the calculate_VSH docstring for more info on this
+        - diffRes = Refer to the calculate_diffRes docstring for more info on this
 
     Args:
         df (pd.DataFrame): dataframe to which add features from and to
@@ -192,7 +226,7 @@ def add_petrophysical_features(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
             df = petro.calculate_RAVG(df)
 
         if "LFI" in petrophysical_features:
-            df = petro.calculate_LFI(df)
+            df = petro.calculate_LFI(df, **kwargs)
 
         if "FI" in petrophysical_features:
             df = petro.calculate_FI(df)
@@ -209,6 +243,9 @@ def add_petrophysical_features(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
         if "VSH" in petrophysical_features:
             df = petro.calculate_VSH(df, **kwargs)
 
+        if "diffRes" in petrophysical_features:
+            df = petro.calculate_diffRes(df, **kwargs)
+
     return df
 
 
@@ -247,18 +284,20 @@ def add_well_metadata(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
     metadata_dict: Dict[str, Dict[str, Any]] = kwargs.get("metadata_dict", None)
     metadata_columns: List[str] = kwargs.get("metadata_columns", None)
     retrieve_from_cdf: bool = kwargs.get("retrieve_from_cdf", metadata_dict is None)
+    client: CogniteClient = kwargs.get("client", None)
 
     if id_column is None:
         raise ValueError("id_column kwarg must be provided")
     if retrieve_from_cdf:
-        client = utilities.get_cognite_client()
+        if client is None:
+            client = utilities.get_cognite_client()
         try:
             metadata_dict = utilities.get_well_metadata(
                 well_names=df[id_column].unique(), client=client
             )
         except Exception as exc:
             raise Exception(
-                "Something failed in the retrieval of the formation tops mapping. "
+                "Something failed in the retrieval of the well metadata. "
                 "Please check the error message below and try again.\n\n"
                 f"{exc}"
             ) from exc
@@ -359,6 +398,8 @@ def add_formations_and_groups(
         str, Dict[str, Union[List[str], List[float]]]
     ] = kwargs.get("formation_tops_mapper", {})
     retrieve_from_cdf: bool = kwargs.get("retrieve_from_cdf", not formation_tops_mapper)
+    add_systems: bool = kwargs.get("add_systems", False)
+    client: CogniteClient = kwargs.get("client", None)
 
     if depth_column not in df.columns:
         raise ValueError(
@@ -369,7 +410,8 @@ def add_formations_and_groups(
 
     well_names = df[id_column].unique()
     if retrieve_from_cdf:
-        client = utilities.get_cognite_client()
+        if client is None:
+            client = utilities.get_cognite_client()
         try:
             formation_tops_mapper = utilities.get_formation_tops(
                 well_names=well_names, client=client
@@ -433,15 +475,7 @@ def add_formations_and_groups(
                 right=False,
                 ordered=False,
             )
-        df_["GROUP"] = utilities.map_formation_and_group(
-            df_["GROUP"].astype(str).apply(utilities.standardize_group_formation_name)
-        )[1]
-
-        df_["FORMATION"] = utilities.map_formation_and_group(
-            df_["FORMATION"]
-            .astype(str)
-            .apply(utilities.standardize_group_formation_name)
-        )[0]
+        df_ = utilities.normalize_group_formation_system(df_, add_systems)
 
     else:
         raise ValueError(
@@ -502,14 +536,15 @@ def add_vertical_depths(
     if id_column is None:
         raise ValueError("No id_column kwarg provided!")
     if retrieve_from_cdf:
-        client = utilities.get_cognite_client()
+        if client is None:
+            client = utilities.get_cognite_client()
         try:
             vertical_depths_mapper = utilities.get_vertical_depths(
                 well_names=df[id_column].unique(), client=client
             )
         except Exception as exc:
             raise Exception(
-                "Something failed in the retrieval of the formation tops mapping. "
+                "Something failed in the retrieval of the vertical depths mapping. "
                 "Please check the error message below and try again.\n\n"
                 f"{exc}"
             ) from exc
@@ -554,6 +589,9 @@ def add_depth_trend(df: pd.DataFrame, **kwargs):
     Keyword Args:
         id_column (str): REQUIRED identifier for the well column in the provided dataframe
             Defaults to None
+        keyword_arguments (dict): REQUIRED dictionary containing any kwargs for the function call.
+            Must contain 'nan_textual_value' & 'nan_numerical_value' key value
+            pairs.
         client (CogniteClient): OPTIONAL client for querying depth trend function from CDF
             Defaults to None
         env (str): OPTIONAL environment where function is hosted (typically test or prod). Defaults to prod
@@ -561,12 +599,6 @@ def add_depth_trend(df: pd.DataFrame, **kwargs):
         return_CI (bool): OPTIONAL whether to return confidence interval to the trends.
         return_file (bool): OPTIONAL whether to return the predictions as a CDF file link.
             Defaults to True.
-        keyword_arguments (dict): OPTIONAL dictionary containing any keywargs for the function call. Eg.
-                keyword_arguments = {
-                    "nan_textual_value": "-9999.0",
-                    "nan_numerical_value": -9999
-                }
-            Defaults to a dictionary as example above.
 
     Returns:
         pd.DataFrame: dataframe with added depth trend and optionally its confidence intervals
@@ -579,29 +611,33 @@ def add_depth_trend(df: pd.DataFrame, **kwargs):
     return_CI: bool = kwargs.get("return_CI", False)
     user_kwargs: Dict[str, Any] = kwargs.get("keyword_arguments", None)
 
-    # Keyword arguments dict has to be initialized if None is given by the user
-    # as the function cannot accept nan in the input.
-    keyword_arguments = {
-        "nan_numerical_value": -9999,
-        "nan_textual_value": "-9999.0",
-    }
-    if user_kwargs is not None:
-        keyword_arguments.update(user_kwargs)
-
     # Validate input parameters
     if id_column is None:
         raise ValueError("id column (well name column) is a required kwarg!")
+    if user_kwargs is None:
+        raise ValueError(
+            "keyword_arguments is a required kwarg! Please refer to the docstring for more information."
+        )
+    else:
+        if "nan_textual_value" not in user_kwargs:
+            raise ValueError(
+                "nan_textual_value is a required kwarg! Please refer to the docstring for more information."
+            )
+        if "nan_numerical_value" not in user_kwargs:
+            raise ValueError(
+                "nan_numerical_value is a required kwarg! Please refer to the docstring for more information."
+            )
     if version is None:
         version = "latest"
         external_id = f"depth_trend-prediction-{env}"
     else:
         external_id = f"depth_trend-prediction-{env}-{version}"
-    df_ = utilities.run_CDF_function(
+    df_ = utilities.run_deployed_model(
         df,
         id_column,
-        keyword_arguments,
-        external_id,
-        client,
+        user_kwargs,
+        cdf_external_id=external_id,
+        client=client,
         return_file=kwargs.get("return_file", True),
     )
     if not return_CI:
