@@ -20,7 +20,7 @@ the Dataset class.
 """
 import os
 import warnings
-from typing import Any, Dict, List, Tuple, Union
+from typing import Any, Dict, List, Optional, Tuple, Union
 
 import joblib
 import numpy as np
@@ -88,11 +88,12 @@ def remove_outliers(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
     curves and corresponding cutoff values are used (if they exist in the
     provided curves list):
 
-        - GR: low cutoff: 0, high cutoff: 250
-        - RMED: high cutoff: 100
-        - RDEP: high cutoff: 100
-        - RSHA: high cutoff: 100
-        - NEU: low cutoff: -0.5, high cutoff: 1 (replaced with np.nan)
+        - GR: low cutoff: 0, high cutoff: 250 gAPI
+        - RMIC: low cutoff: 0, high cutoff: 100 above which set to 100
+        - RMED: low cutoff: 0, high cutoff: 100 above which set to 100
+        - RDEP: low cutoff: 0, high cutoff: 100 above which set to 100
+        - RSHA: low cutoff: 0, high cutoff: 100 above which set to 100
+        - NEU: low cutoff: -0.5, high cutoff: 1 (replaced with np.nan) v/v
         - PEF: high cutoff: 10 (replaced with np.nan)
 
     If not otherwise specified above, values above/below the cutoffs are replaced
@@ -128,10 +129,10 @@ def remove_outliers(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
                     )
                 df.GR = df.GR.clip(lower=0, upper=250)
 
-            for resistivity in ["RSHA", "RMED", "RDEP"]:
+            for resistivity in ["RMIC", "RSHA", "RMED", "RDEP"]:
                 if resistivity in outlier_curves:
-                    outliers = df[df[resistivity] > 100]
-                    if len(outliers) / len_df > th:
+                    outliers = df[resistivity] > 100
+                    if outliers.sum() / len_df > th:
                         warnings.warn(
                             f"{resistivity} has more than {th*100}% of its values higher"
                             " than 100. Note: This column names is the name after it has"
@@ -139,27 +140,30 @@ def remove_outliers(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
                             " be another column from your original data that triggered"
                             " this warning and instead was mapped to the name printed above."
                         )
-                    df.loc[outliers.index, resistivity] = 100
+                    df.loc[outliers, resistivity] = 100
+                    # Remove negative values
+                    df.loc[df[resistivity] < 0, resistivity] = np.nan
 
             if "NEU" in outlier_curves:
-                outliers = df[(df.NEU < -0.5) | (df.NEU > 1)]
-                ratio = len(outliers) / len_df
+                outliers = (df.NEU < -0.5) | (df.NEU > 1)
+                ratio = outliers.sum() / len_df
                 if ratio > th:
                     warnings.warn(
                         f"NEU has {ratio*100}% of its values higher than 1"
                         " or lower than -0.5. The threshold for this warning is"
                         f" {th*100}%"
                     )
-                df.loc[df.NEU > 1, "NEU"] = np.nan
-                df.loc[df.NEU < -0.5, "NEU"] = np.nan
+                df.loc[outliers, "NEU"] = np.nan
 
             if "PEF" in outlier_curves:
-                outliers = df[df.PEF > 10]
-                if len(outliers) / len_df > th:
+                outliers = df.PEF > 10
+                if outliers.sum() / len_df > th:
                     warnings.warn(
                         f"PEF has more than {th*100}% of its values higher than 10"
                     )
-                df.loc[df.PEF > 10, "PEF"] = np.nan
+                df.loc[outliers, "PEF"] = np.nan
+                # Remove negative PEF values
+                df.loc[df.PEF < 0, "PEF"] = np.nan
 
     return df
 
@@ -260,12 +264,12 @@ def encode_columns(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
             provided in the kwargs.
         - GROUP column - categories are encoded using the groups_map
             provided in the kwargs
-        - lsuName column - categories are encoded using the groups_map
+        - SYSTEM column - categories are encoded using the systems_map
             provided in the kwargs
 
-    Note: All names are standardized prior to mapping using the utility function
-        standardize_group_formation_name and all categories that weren't mapped
-        are encoded with -1.
+    Note: All names are normalized prior to mapping using the utility function
+        normalize_group_formation_system and all categories that weren't mapped
+        are encoded with missing_encoding_value.
 
     Args:
         df (pd.DataFrame): dataframe to which apply encoding of categorical variables
@@ -275,40 +279,96 @@ def encode_columns(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
             encoded. If no columns are passed the get_col_types utility function
             is used to determine the categorical columns
         formations_map (dict): A mapping dictionary mapping formation names to
-            corresponding integers. Defaults to an empty dictionary (ie no encoding).
+            corresponding numbers. Defaults to an empty dictionary (ie no encoding).
         groups_map (dict): A mapping dictionary mapping group names to
-            corresponding integers. Defaults to an empty dictionary (ie no encoding).
-        missing_encoding_value (int): The value to fill encode categories for which
-            no match was found in the provided mappings. Defaults to -1.
+            corresponding numbers. Defaults to an empty dictionary (ie no encoding).
+        systems_map (dict): A mapping dictionary mapping system names to
+            corresponding numbers. Defaults to an empty dictionary (ie no encoding).
+        missing_encoding_value (float): The value to fill encode categories for which
+            no match was found in the provided mappings. Defaults to -1.0.
 
     Returns:
         pd.DataFrame: dataframe with categorical columns encoded
     """
     columns_to_encode: List[str] = kwargs.get("columns_to_encode", None)
-    formations_map: Dict[str, int] = kwargs.get("formations_map", None)
-    groups_map: Dict[str, int] = kwargs.get("groups_map", None)
-    missing_encoding_value: int = kwargs.get("missing_encoding_value", -1)
+    formations_map: Dict[str, float] = kwargs.get("formations_map", None)
+    groups_map: Dict[str, float] = kwargs.get("groups_map", None)
+    systems_map: Dict[str, float] = kwargs.get("systems_map", None)
+    missing_encoding_value: float = kwargs.get("missing_encoding_value", -1.0)
 
     if columns_to_encode is None:
         _, columns_to_encode = utilities.get_col_types(df, columns_to_encode)
+
+    # Normalize before mapping
+    df = utilities.normalize_group_formation_system(
+        df, add_systems="SYSTEM" in columns_to_encode
+    )
     if "FORMATION" in columns_to_encode and formations_map is not None:
-        df["FORMATION"] = df["FORMATION"].apply(
-            utilities.standardize_group_formation_name
-        )
         df["FORMATION"] = df["FORMATION"].map(formations_map)
         df["FORMATION"] = df["FORMATION"].fillna(missing_encoding_value)
+        df["FORMATION"] = df["FORMATION"].astype(float)
     if "GROUP" in columns_to_encode and groups_map is not None:
-        df["GROUP"] = df["GROUP"].apply(utilities.standardize_group_formation_name)
         df["GROUP"] = df["GROUP"].map(groups_map)
         df["GROUP"] = df["GROUP"].fillna(missing_encoding_value)
-    if "lsuName" in columns_to_encode and groups_map is not None:
-        df["lsuName"] = df["lsuName"].apply(utilities.standardize_group_formation_name)
-        df["lsuName"] = df["lsuName"].map(groups_map)
-        df["lsuName"] = df["lsuName"].fillna(missing_encoding_value)
+        df["GROUP"] = df["GROUP"].astype(float)
+    if "SYSTEM" in columns_to_encode and systems_map is not None:
+        df["SYSTEM"] = df["SYSTEM"].map(systems_map)
+        df["SYSTEM"] = df["SYSTEM"].fillna(missing_encoding_value)
+        df["SYSTEM"] = df["SYSTEM"].astype(float)
 
     return df
 
 
+def onehot_encode_columns(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
+    """
+    Onehot encodes all columns provided via the columns_to_onehot_encode kwarg.
+
+    Note: NaNs are not encoded. They are separated into an individual NaN column
+        and that column is automatically dropped.
+
+    Args:
+        df (pd.DataFrame): dataframe to which apply onehot encoding of categorical variables
+
+    Keyword Args:
+        columns_to_onehot_encode (list): which columns to encode. Default to no columns being
+            encoded. If no columns are passed, this method raises a warning and
+            does nothing.
+        categories_to_keep (list): A list of categories to keep. If None, all categories
+            are kept. The categories are all prefixed with their corresponding column name and an
+            underscore. For example, if the column name is "GROUP" and the category is "CROMER KNOLL GP", the
+            category name would be "GROUP_CROMER KNOLL GP".
+
+    Returns:
+        pd.DataFrame: dataframe with categorical columns encoded
+    """
+    columns_to_onehot_encode: List[str] = kwargs.get("columns_to_onehot_encode", None)
+    categories_to_keep: List[str] = kwargs.get("categories_to_keep", None)
+    if columns_to_onehot_encode is not None:
+        encoded = pd.get_dummies(
+            df[columns_to_onehot_encode],
+            columns=columns_to_onehot_encode,
+            dummy_na=False,
+        )
+        cols_to_keep = (
+            encoded.columns if categories_to_keep is None else categories_to_keep
+        )
+        for col in cols_to_keep:
+            if col not in encoded.columns:
+                warnings.warn(
+                    f"{col} could not be found after one hot encoding. "
+                    "Initialising it with all np.nans! Make sure to check the "
+                    "categories_to_keep argument!"
+                )
+                encoded[col] = np.nan
+        df = pd.concat([df, encoded[cols_to_keep]], axis=1)
+        return df, {"onehot_columns": cols_to_keep}
+    else:
+        warnings.warn(
+            "No columns to onehot encode provided. Doing nothing.", UserWarning
+        )
+        return df
+
+
 def select_columns(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
     """
     Returns a dataframe with only curves chosen by user, filtered from the original dataframe
@@ -362,6 +422,36 @@ def drop_columns(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
     return df
 
 
+def fill_columns(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
+    """
+    Fills missing values in the dataframe according to the fill mapping provided.
+
+    Args:
+        df (pd.DataFrame): dataframe to fill
+        columns_to_fill (dict): which columns to fill where the keys
+            are the main columns to be filled in with values from the columns
+            listed in the values. Defaults to None.
+        missing_value (float): The value that is used to represent missing values. Defaults to np.nan
+
+    Returns:
+        pd.DataFrame: dataframe with columns filled according to the provided mapping
+    """
+    columns_to_fill: Dict[str, List[str]] = kwargs.get("columns_to_fill", None)
+    missing_value: float = kwargs.get("missing_value", None)
+
+    if missing_value is None:
+        missing_value = np.nan
+
+    if columns_to_fill is not None:
+        for main_column, columns_to_fill_with in columns_to_fill.items():
+            df[main_column] = df[main_column].replace(missing_value, np.nan)
+            for column_to_fill_with in columns_to_fill_with:
+                df[main_column] = df[main_column].fillna(df[column_to_fill_with])
+            df[main_column] = df[main_column].fillna(missing_value)
+
+    return df
+
+
 def normalize_curves(
     df: pd.DataFrame, **kwargs
 ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, Dict[str, Any]]]:
@@ -465,9 +555,9 @@ def normalize_curves(
             # normalize all other wells using key well as reference
             df.loc[:, c] = df.apply(
                 lambda x: utilities.normalize(
-                    x[c],
-                    key_well["ref_low"],
-                    key_well["ref_high"],
+                    x[c],  # noqa: B023
+                    key_well["ref_low"],  # noqa: B023
+                    key_well["ref_high"],  # noqa: B023
                     x["low_p"],
                     x["high_p"],
                 ),
@@ -710,7 +800,7 @@ def apply_calibration(
     mode: str,
     id_column: str,
     distance_thres: float = 9999.0,
-    calib_map: pd.DataFrame = None,
+    calib_map: Optional[pd.DataFrame] = None,
     standardize_level_names: bool = True,
 ) -> pd.DataFrame:
     """
