@@ -27,6 +27,18 @@ TEST_DF = pd.DataFrame(
             87126: 12.25,
             87127: 12.25,
         },
+        "CALI": {
+            12535: 12.25,
+            12536: 12.25,
+            30344: 17.5,
+            30345: 17.5,
+            48749: nan,
+            48750: nan,
+            69538: 12.25,
+            69539: 12.25,
+            87126: 12.25,
+            87127: 12.25,
+        },
         "DENC": {
             12535: nan,
             12536: nan,
@@ -99,7 +111,7 @@ TEST_DF = pd.DataFrame(
             87126: 0.7484,
             87127: 0.7484,
         },
-        "lsuName": {
+        "GROUP": {
             12535: "ROGALAND",
             12536: "ROGALAND",
             30344: "HORDALAND",
@@ -208,7 +220,7 @@ TRAIN_DF = pd.DataFrame(
             87126: 0.7484,
             87127: 0.7484,
         },
-        "lsuName": {
+        "GROUP": {
             12535: "ROGALAND",
             12536: "ROGALAND",
             30344: "HORDALAND",
@@ -344,29 +356,29 @@ PREPROCESSED_DF = pd.DataFrame(
             87127: 1.7705087834084305,
         },
         "GR_window_max": {
-            12535: 99.78,
+            12535: 100.49,
             12536: 100.49,
             30344: 108.207,
             30345: 108.207,
-            69538: 55.46915,
+            69538: 60.19225,
             69539: 60.19225,
-            87126: 56.51378,
+            87126: 57.95339,
             87127: 57.95339,
         },
         "GR_window_mean": {
-            12535: 99.78,
+            12535: 100.13499999999999,
             12536: 100.13499999999999,
-            30344: 108.207,
+            30344: 108.1199,
             30345: 108.1199,
-            69538: 55.46915,
+            69538: 57.8307,
             69539: 57.8307,
-            87126: 56.51378,
+            87126: 57.233585,
             87127: 57.233585,
         },
         "GR_window_min": {
             12535: 99.78,
             12536: 99.78,
-            30344: 108.207,
+            30344: 108.0328,
             30345: 108.0328,
             69538: 55.46915,
             69539: 55.46915,
@@ -426,25 +438,25 @@ PREPROCESSED_DF = pd.DataFrame(
         "RDEP_window_max": {
             12535: 0.77,
             12536: 0.77,
-            30344: 0.4615473,
+            30344: 0.5001374,
             30345: 0.5001374,
-            69538: 0.525528,
+            69538: 0.5493600000000001,
             69539: 0.5493600000000001,
             87126: 0.7484,
             87127: 0.7484,
         },
         "RDEP_window_mean": {
-            12535: 0.77,
+            12535: 0.75,
             12536: 0.75,
-            30344: 0.4615473,
+            30344: 0.48084235,
             30345: 0.48084235,
-            69538: 0.525528,
+            69538: 0.537444,
             69539: 0.537444,
             87126: 0.7484,
             87127: 0.7484,
         },
         "RDEP_window_min": {
-            12535: 0.77,
+            12535: 0.73,
             12536: 0.73,
             30344: 0.4615473,
             30345: 0.4615473,
@@ -486,7 +498,7 @@ PREPROCESSED_DF = pd.DataFrame(
         "RMED_window_max": {
             12535: 0.7,
             12536: 0.7,
-            30344: 0.4679713,
+            30344: 0.5017649000000001,
             30345: 0.5017649000000001,
             69538: 0.71134,
             69539: 0.71134,
@@ -494,34 +506,34 @@ PREPROCESSED_DF = pd.DataFrame(
             87127: 0.7484,
         },
         "RMED_window_mean": {
-            12535: 0.7,
+            12535: 0.69000005,
             12536: 0.69000005,
-            30344: 0.4679713,
+            30344: 0.4848681,
             30345: 0.4848681,
-            69538: 0.71134,
+            69538: 0.69121,
             69539: 0.69121,
             87126: 0.7484,
             87127: 0.7484,
         },
         "RMED_window_min": {
-            12535: 0.7,
+            12535: 0.6800001,
             12536: 0.6800001,
             30344: 0.4679713,
             30345: 0.4679713,
-            69538: 0.71134,
+            69538: 0.67108,
             69539: 0.67108,
             87126: 0.7484,
             87127: 0.7484,
         },
-        "lsuName": {
-            12535: 21,
-            12536: 21,
-            30344: 14,
-            30345: 14,
-            69538: 14,
-            69539: 14,
-            87126: 14,
-            87127: 14,
+        "GROUP": {
+            12535: 63.0,
+            12536: 63.0,
+            30344: 53.0,
+            30345: 53.0,
+            69538: 53.0,
+            69539: 53.0,
+            87126: 53.0,
+            87127: 53.0,
         },
         "well_name": {
             12535: "30/6-26",
@@ -668,8 +680,24 @@ FORMATION_DF = pd.DataFrame(
             659959: "ROGALAND GP",
             660176: "SHETLAND GP",
             660195: "STATFJORD GP",
-            660858: "UNKNOWN GP",
-            662138: "UNKNOWN GP",
+            660858: "HEGRE GP",
+            662138: "HEGRE GP",
+        },
+        "SYSTEM": {
+            650372: "NORDLANDS SY",
+            651258: "NORDLANDS SY",
+            651698: "HORDALANDS SY",
+            654933: "HORDALANDS SY",
+            656475: "HORDALANDS SY",
+            658201: "ROGALANDS SY",
+            658496: "ROGALANDS SY",
+            658745: "ROGALANDS SY",
+            658778: "ROGALANDS SY",
+            659959: "ROGALANDS SY",
+            660176: "CRETACEOUS SY",
+            660195: "TRIASSIC SY",
+            660858: "TRIASSIC SY",
+            662138: "TRIASSIC SY",
         },
         "well_name": {
             650372: "25/10-10",
