@@ -4,13 +4,15 @@ import os
 import re
 import warnings
 from functools import lru_cache
-from typing import Any, Callable, Dict, List, Set, Tuple, Union
+from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
 
 import numpy as np
 import numpy.typing as npt
 import pandas as pd
+import requests
+from cognite.client import ClientConfig, CogniteClient
+from cognite.client.credentials import APIKey
 from cognite.client.data_classes import Asset, AssetList
-from cognite.experimental import CogniteClient
 from numpy import float64
 
 import akerbp.mlpet.data
@@ -64,9 +66,9 @@ def readPickle(path):
     return data
 
 
-def map_formation_and_group(
-    form_or_group: pd.Series, MissingValue: Union[float, str] = np.nan
-) -> Tuple[Union[float, str], Union[float, str]]:
+def map_formation_group_system(
+    form_or_group_or_system: pd.Series, MissingValue: Union[float, str] = np.nan
+) -> Tuple[Union[float, str], Union[float, str], Union[float, str]]:
     """
     A helper function for retrieving the formation and group of a standardised
     formation/group based on mlpet's NPD pickle mapper.
@@ -80,27 +82,100 @@ def map_formation_and_group(
         tuple(pd.Series): Returns a formation and group series respectively corresponding
             to the input string series
     """
-    with importlib.resources.path(akerbp.mlpet.data, "npd_fm_gp_key_dic.pcl") as path:
+    with importlib.resources.path(
+        akerbp.mlpet.data, "npd_fm_gp_sy_key_dic.pcl"
+    ) as path:
         dic_names = readPickle(path)
 
     mapping = {}
-    for item in form_or_group.unique():
-        form, group = MissingValue, MissingValue
+    for item in form_or_group_or_system.unique():
+        form, group, siistem = MissingValue, MissingValue, MissingValue
         try:
             dic = dic_names[item]
+            name = dic["NAME"]
+            parent = dic["PARENT"]
+            grandparent = dic["GRANDPARENT"]
             if dic["LEVEL"] == "FORMATION":
-                form = dic["NAME"]
-                if " GP" in dic["PARENT"]:
-                    group = dic["PARENT"]
+                form = name
+                group = parent
+                siistem = grandparent
             elif dic["LEVEL"] == "GROUP":
-                group = dic["NAME"]
-        except KeyError:
+                group = name
+                siistem = grandparent
+            elif dic["LEVEL"] == "SYSTEM":
+                siistem = name
+        except (KeyError, ValueError):
             pass
-        mapping[item] = (form, group)
+        form, group, siistem = pd.Series([form, group, siistem]).fillna(MissingValue)
+        mapping[item] = (form, group, siistem)
 
-    form, group = zip(*form_or_group.map(mapping))
+    form, group, siistem = zip(*form_or_group_or_system.map(mapping))
 
-    return form, group
+    return form, group, siistem
+
+
+def normalize_group_formation_system(
+    df_: pd.DataFrame, add_systems=False
+) -> pd.DataFrame:
+    """
+    A helper function to normalize a GROUP and FORMATION column using mlpet's
+    builtin NPD mapper. A SYSTEM column can also optionally be added by setting
+    add_systems to True.
+
+    Args:
+        add_systems (bool): Whether to add a SYSTEM column to the dataframe based
+            on the GROUP and FORMATION columns mapping
+        df_ (pd.DataFrame): The dataframe containing both the GROUP and FORMATION
+            columns to be normalized
+
+    Returns:
+        pd.DataFrame: The dataframe with the normalization applied
+    """
+    if "FORMATION" in df_.columns:
+        (
+            df_["FORMATION"],
+            df_["GROUP_MAPPED"],
+            df_["SYSTEM_MAPPED_FM"],
+        ) = map_formation_group_system(
+            df_["FORMATION"].astype(str).apply(standardize_group_formation_name)
+        )
+    if "GROUP" in df_.columns:
+        _, df_["GROUP"], df_["SYSTEM_MAPPED_GP"] = map_formation_group_system(
+            df_["GROUP"].astype(str).apply(standardize_group_formation_name)
+        )
+        # NOTE: If GROUP is provided to this function and some filling is done,
+        # ensure that the FORMATION column is adjusted as well. If a group is
+        # inferred, a FORMATION should be set to UNKNOWN unless it was already
+        # provided in the FORMATION column sent to this function
+
+        # If formation is aviailable enrich the group column with it
+        if "FORMATION" in df_.columns:
+            # Special fix for increasing data quality of GROUP column
+            fill_index = df_["FORMATION"].ne("UNKNOWN FM") & df_["GROUP"].eq(
+                "UNKNOWN GP"
+            )
+            if fill_index.any():
+                df_.loc[fill_index, "GROUP"] = df_.loc[fill_index, "GROUP_MAPPED"]
+
+            # If no group information is available, use the formation mapped groups
+            df_["GROUP"] = df_["GROUP"].fillna(df_["GROUP_MAPPED"])
+    if add_systems:
+        if "SYSTEM" in df_.columns:
+            _, _, df_["SYSTEM"] = map_formation_group_system(
+                df_["SYSTEM"].astype(str).apply(standardize_group_formation_name)
+            )
+        else:
+            df_["SYSTEM"] = np.nan
+        for col in ["SYSTEM_MAPPED_GP", "SYSTEM_MAPPED_FM"]:
+            if col in df_.columns:
+                df_["SYSTEM"] = df_["SYSTEM"].fillna(df_[col])
+
+    df_ = df_.drop(
+        columns=["GROUP_MAPPED", "SYSTEM_MAPPED_FM", "SYSTEM_MAPPED_GP"],
+        errors="ignore",
+    )
+
+    return df_
 
 
 def standardize_group_formation_name(name: Union[str, Any]) -> Union[str, Any]:
@@ -136,14 +211,21 @@ def standardize_group_formation_name(name: Union[str, Any]) -> Union[str, Any]:
     if name == "NAN":
         return np.nan
     # GPs & FMs with no definition leave as is
-    if name in [
+    unknown = [
         "NO FORMAL NAME",
         "NO GROUP DEFINED",
         "UNDEFINED",
         "UNDIFFERENTIATED",
         "UNKNOWN",
-    ]:
-        return "UNKNOWN"
+    ]
+    for item in unknown:
+        if item in name.upper():
+            return "UNKNOWN"
+
+    # Remove known prefixes
+    known_prefixes = ["LOWER", "UPPER"]
+    for prefix in known_prefixes:
+        name = name.replace(prefix, "").strip()
 
     # Then perform standardization
     if "INTRA" in name:
@@ -207,7 +289,7 @@ def standardize_curve_names(df: pd.DataFrame, mapper: Dict[str, str]) -> pd.Data
 
 
 def get_col_types(
-    df: pd.DataFrame, categorical_curves: List[str] = None, warn: bool = True
+    df: pd.DataFrame, categorical_curves: Optional[List[str]] = None, warn: bool = True
 ) -> Tuple[List[str], List[str]]:
     """
     Returns lists of numerical and categorical columns
@@ -291,7 +373,7 @@ def df_split_train_test(
     df: pd.DataFrame,
     id_column: str,
     test_size: float = 0.2,
-    test_wells: List[str] = None,
+    test_wells: Optional[List[str]] = None,
     **kwargs,
 ) -> Tuple[pd.DataFrame, pd.DataFrame, List[str]]:
     """
@@ -546,8 +628,8 @@ def get_well_metadata(  # noqa: C901
     for well in well_names:
         try:
             # First try list search
-            asset: Union[AssetList, List[Asset], Asset] = client.assets.list(name=well)
-            if len(asset) == 0:
+            assets: Union[AssetList, List[Asset]] = client.assets.list(name=well)
+            if len(assets) == 0:
                 # If first attempt failed use fuzzy search to retrieve proper
                 # well name. Find best match based on alphanumeric equality
                 wellName = _find_best_match(
@@ -561,16 +643,18 @@ def get_well_metadata(  # noqa: C901
                     "metadata from that asset!"
                 )
                 # Then retrieve asset using list API
-                asset = client.assets.list(name=wellName, metadata={"type": "Wellbore"})
-            if len(asset) > 1:
+                assets = client.assets.list(
+                    name=wellName, metadata={"type": "Wellbore"}
+                )
+            if len(assets) > 1:
                 # Sort by time with first element being most recent
-                asset = sorted(asset, key=lambda x: x.last_updated_time)
+                assets = sorted(assets, key=lambda x: x.last_updated_time)
                 # Some wells are stored several times as assets??
-                # In this case find merge them all together to retrieve as much
-                # metadata as possible. Where a me
-                series_meta = _merge_assets(asset)
+                # In this case merge them all together to retrieve as much
+                # metadata as possible.
+                series_meta = _merge_assets(assets)
             else:
-                asset = asset[0]
+                asset = assets[0]
                 series_meta = asset.to_pandas().squeeze()
                 series_meta.name = asset.name
         except IndexError:
@@ -628,7 +712,7 @@ def get_well_metadata(  # noqa: C901
             # Respect order of similar key mapping
             highest_rank_key = idx[np.argmin([mapping.index(x) for x in idx])]
             values = cdf_meta.loc[idx].apply(
-                lambda x: _apply_function(x, highest_rank_key), axis=0
+                lambda x: _apply_function(x, highest_rank_key), axis=0  # noqa: B023
             )
             values.name = mapping_name
         cdf_meta = cdf_meta.loc[~cdf_meta.index.isin(idx)]
@@ -717,8 +801,8 @@ def get_formation_tops(  # noqa: C901
                         rows.loc[i, "Lithostrat"] = standardize_group_formation_name(
                             r.Lithostrat
                         )
-                    rows["FORMATION"], rows["GROUP"] = map_formation_and_group(
-                        form_or_group=rows.Lithostrat, MissingValue="UNKNOWN"
+                    rows["FORMATION"], rows["GROUP"], _ = map_formation_group_system(
+                        form_or_group_or_system=rows.Lithostrat, MissingValue="UNKNOWN"
                     )
 
                     # For formation without a group recording to mapping
@@ -861,7 +945,7 @@ def get_vertical_depths(
             )
             continue
         well_df_discrete = client.sequences.data.retrieve_dataframe(
-            id=well_data_cdf[0].id, start=None, end=None
+            id=well_data_cdf[0].id, start=0, end=None
         )
         if len(well_df_discrete) == 0:
             warnings.warn(
@@ -891,7 +975,7 @@ def get_calibration_map(
     location_curves: List[str],
     mode: str,
     id_column: str,
-    levels: List[str] = None,
+    levels: Optional[List[str]] = None,
     standardize_level_names: bool = True,
 ) -> Dict[str, pd.DataFrame]:
     """
@@ -961,7 +1045,7 @@ def get_calibration_values(
     mode: str,
     id_column: str,
     distance_thres: float = 99999.0,
-    calibration_map: pd.DataFrame = None,
+    calibration_map: Optional[pd.DataFrame] = None,
     standardize_level_names: bool = True,
 ) -> Dict[str, pd.DataFrame]:
     """
@@ -1081,7 +1165,7 @@ def get_calibration_values(
     return well_values
 
 
-def calculate_sampling_rate(array, max_sampling_rate=1):
+def calculate_sampling_rate(array: pd.Series, max_sampling_rate=1):
     """
     Calculates the sampling rate of an array by calculating the weighed
     average diff between the array's values.
@@ -1090,17 +1174,17 @@ def calculate_sampling_rate(array, max_sampling_rate=1):
         array (pd.Series): The array for which the sampling rate should be calculated
         max_sampling_rate: The maximum acceptable sampling rate above which the
             the calculated sampling rates should not be included in the weighted
-            average calculation (defined in samples/unit length e.g. m). Defaults
-            to max 1 sample per m (where m is the assumed unit of the provided array)
+            average calculation (defined in unit length/sample e.g. m). Defaults
+            to max 1 m per sample (where m is the assumed unit of the provided array)
     """
-    if array.empty or np.isnan(array).all():
+    if array.empty or array.isna().all():
         raise ValueError(
             "The provided array is empty or contains only NaNs! Cannot calculate sampling rate!"
         )
     diffs = pd.Series(np.diff(array.values)).value_counts(normalize=True)
     # Ensure big holes in the index don't affect the weighted average
     # Asumming 1 is a good enough threshold for now
-    diffs.loc[diffs.index.to_series().gt(max_sampling_rate)] = np.nan
+    diffs.loc[diffs.index.to_series().abs().gt(max_sampling_rate)] = np.nan
     sampling_rate = (diffs * diffs.index).sum()
     return sampling_rate
 
@@ -1202,7 +1286,7 @@ def estimate_parameter(
 
 
 def get_cognite_client(
-    cognite_api_key: str = None,
+    cognite_api_key: Optional[str] = None,
     cognite_project_id: str = "akbp-subsurface",
     cognite_client_id: str = "mlpet",
 ) -> CogniteClient:
@@ -1217,60 +1301,65 @@ def get_cognite_client(
         Returns:
             CogniteClient: The created CogniteClient instance
     """
+    e = None
     if cognite_api_key is None:
         try:
             cognite_api_key = os.environ["COGNITE_API_KEY"]
-        except KeyError as e:
-            raise ValueError(
-                "No API key provided and no environment variable "
-                "COGNITE_API_KEY found!"
-            ) from e
-    return CogniteClient(
-        api_key=cognite_api_key,
-        project=cognite_project_id,
+        except KeyError:
+            try:
+                from akerbp.mlops.core import config
+
+                cognite_api_key = config.api_keys["data"]
+            except ImportError:
+                # akerbp.mlops is not installed so ignore
+                pass
+            except KeyError:
+                # Could not find the data key in config try to use whichever
+                # key is available
+                v = config.api_keys.values()
+                if v:
+                    cognite_api_key = v[0]
+            except Exception as exc:
+                e = exc
+
+    if cognite_api_key is None:
+        raise ValueError(
+            "No API key provided, none could be imported from akerbp.mlops and no environment variable "
+            "COGNITE_API_KEY was found!"
+        ) from e
+
+    # Assuming COGNITE_API_KEY is set in the environment
+    credentials = APIKey(cognite_api_key)
+    client_config = ClientConfig(
         client_name=cognite_client_id,
+        project=cognite_project_id,
+        credentials=credentials,
     )
+    return CogniteClient(client_config)
 
 
-def run_CDF_function(
+def run_deployed_model(
     df: pd.DataFrame,
     id_column: str,
     keyword_arguments: Dict[str, Any],
-    external_id: str,
-    client: CogniteClient = None,
-    return_file: bool = True,
+    kubeflow_model_url: Optional[str] = None,
+    header: Optional[Dict[Any, Any]] = None,
+    cdf_external_id: Optional[str] = None,
+    **kwargs,
 ) -> pd.DataFrame:
     """
-    Helper function to run a CDF function.
+    Runs a deployed model on a dataframe.
 
-        Args:
-            df (pd.DataFrame): The dataframe to run the function on
-            id_column (str): The column to use as the ID for the function
-            keyword_arguments (Dict[str, Any]): The keyword arguments to use for the function
-            external_id (str): The external ID to use for the function
-            client (CogniteClient): The CogniteClient instance to use for the function
-            return_file (bool): Whether to return the predictions as a CDF file link.
-                Defaults to True.
+    Args:
+        df (pd.DataFrame): The dataframe to run the model on
+        id_column (str): The column in the dataframe that contains the well names
+        kubeflow_model_url (str): The URL to the deployed model
+        cdf_external_id (str): The external ID of the Cognite Data Fusion function
+        keyword_arguments (Dict[str, Any]): Keyword arguments to pass to the model
 
-        Returns:
-            pd.DataFrame: The dataframe with the results of the function
+    Returns:
+        pd.DataFrame: The result of the model run
     """
-    # Create the CogniteClient instance
-    if client is None:
-        client = get_cognite_client()
-    function = client.functions.retrieve(external_id=external_id)
-    if function is None:
-        raise ValueError(
-            f"No function with external ID {external_id} found in Cognite!"
-        )
-    required_input = eval(function.to_pandas().loc["required_input", "value"])
-    optional_input = eval(function.to_pandas().loc["optional_input", "value"])
-    if not np.all([c in df.columns for c in required_input]):
-        raise ValueError(
-            f"The dataframe provided does not contain all required input"
-            f"for the function ({required_input})"
-        )
-    input_curves = [c for c in required_input + optional_input if c in df.columns]
     # We check that the input curves to the function will not have nans before sending them to the function
     df_ = df.copy()
     df_ = preprocessors.fillna_with_fillers(
@@ -1288,30 +1377,30 @@ def run_CDF_function(
         "input": [
             {
                 "well": k,
-                "input_logs": v[input_curves].to_dict(orient="list"),
+                "input_logs": v.drop(columns=[id_column]).to_dict(orient="list"),
                 "keyword_arguments": keyword_arguments,
             }
             for k, v in df_.groupby(id_column)
-        ],
-        "return_file": return_file,
+        ]
     }
-    results = function.call(in_data).get_response()
-    if results["status"] == "error":
-        raise ValueError(f"Function returned an error: {results['error message']}")
-    # get function results to df
-    output_curves = eval(function.to_pandas().loc["output_curves", "value"])
-    df_[output_curves] = np.nan
-    if return_file:
-        predictions = json.loads(
-            client.files.download_bytes(external_id=results["prediction_file"])
-        )
+
+    if cdf_external_id is not None:
+        predictions = run_CDF_function(in_data, cdf_external_id, **kwargs)
+    elif kubeflow_model_url is not None:
+        if header is not None:
+            predictions = run_kubeflow_model(
+                in_data, kubeflow_model_url, header=header, **kwargs
+            )
+        else:
+            predictions = run_kubeflow_model(in_data, kubeflow_model_url, **kwargs)
     else:
-        predictions = results["prediction"]
+        raise ValueError(
+            "Neither a Cognite Data Fusion function nor a Kubeflow model URL was provided!"
+        )
+
     for well in predictions:
         pred = pd.DataFrame.from_dict(well)
-        df_.loc[df_[id_column] == well["well"], output_curves] = pred[
-            output_curves
-        ].values
+        df_.loc[df_[id_column] == well["well"], pred.columns] = pred.values
 
     # Fill imputed values back to original
     df_ = preprocessors.set_as_nan(
@@ -1321,3 +1410,124 @@ def run_CDF_function(
     )
 
     return df_
+
+
+def run_CDF_function(
+    in_data: Dict[str, Any],
+    cdf_external_id: str,
+    client: Optional[CogniteClient] = None,
+    return_file: bool = True,
+    **kwargs,
+) -> pd.DataFrame:
+    """
+    Helper function to run a CDF function.
+
+        Args:
+            in_data: The payload to send to the CDF function
+            external_id (str): The external ID to use for the function
+            client (CogniteClient): The CogniteClient instance to use for the function
+            return_file (bool): Whether to return the predictions as a CDF file link.
+                Defaults to True.
+
+        Returns:
+            pd.DataFrame: The dataframe with the results of the function
+    """
+    # Create the CogniteClient instance
+    if client is None:
+        client = get_cognite_client()
+    function = client.functions.retrieve(external_id=cdf_external_id)
+    if function is None:
+        raise ValueError(
+            f"No function with external ID {cdf_external_id} found in Cognite!"
+        )
+    in_data["return_file"] = return_file
+
+    results = function.call(in_data).get_response()
+    if results["status"] == "error":
+        keys = [k for k in list(results) if "message" in k.lower()]
+        if keys:
+            raise ValueError(f"Function returned an error: {results[keys[0]]}")
+        else:
+            raise ValueError(f"Function returned an error: {results}")
+    if return_file:
+        predictions = json.loads(
+            client.files.download_bytes(external_id=results["prediction_file"])
+        )
+    else:
+        predictions = results["prediction"]
+
+    return predictions
+
+
+def run_kubeflow_model(
+    in_data: Dict[str, Any],
+    model_url: str,
+    header: Optional[Dict[Any, Any]] = None,
+    session: Optional[requests.Session] = None,
+    **kwargs,
+) -> pd.DataFrame:
+    """
+    Helper function to send a http(s) request to a model hosted on a (kubeflow) cluster.
+    Note: this function assumes requests are sent to url on the form
+    'http(s)://(...)/model_name/endpoint',
+    where 'endpoint' can be either 'metadata' or 'predict'.
+
+        Args:
+            in_data: The payload to send to the kubeflow model
+            model_url (str): The URL to the kubeflow model
+            session (requests.Session): The requests session to use for sending the request
+        Returns:
+            pd.DataFrame: The dataframe with the results of the function
+        Raises:
+            ValueError: If an invalid endpoint is requested, i.e. other than 'predict' and 'metadata'
+            ValueError: If request exits with a status code other than 200
+    """
+    model_name = model_url.split("/")[-2]
+    endpoint = model_url.split("/")[-1]
+    if endpoint == "metadata":
+        method = "get"
+    elif endpoint == "predict":
+        method = "post"
+    else:
+        raise ValueError(
+            f"Invalid request: Model {model_name} does not have a '{endpoint}' endpoint"
+        )
+    # Create a session to send the request
+    if session is None:
+        session = requests.Session()
+
+    # Send the request
+    if method == "post":
+        try:
+            if header is not None:
+                response = session.post(
+                    url=model_url,
+                    json=in_data,
+                    headers=header,
+                )
+            else:
+                response = session.post(model_url, json=in_data)
+        except requests.exceptions.ConnectionError as e:
+            raise ValueError(f"Failed to connect to endpoint '{model_url}'") from e
+    else:
+        try:
+            if header is not None:
+                response = session.get(
+                    url=model_url,
+                    headers=header,
+                )
+            else:
+                response = session.get(model_url)
+        except requests.exceptions.ConnectionError as e:
+            raise ValueError(f"Failed to connect to endpoint '{model_url}'") from e
+
+    # Check that the response is OK
+    if response.status_code != 200:
+        raise ValueError(
+            f"Request to {model_url} failed with status code {response.status_code}! See response for details: {response.text}"
+        )
+    # Get the predictions/metadata
+    if endpoint == "metadata":
+        return response.json()
+    else:
+        return response.json()["prediction"]
